@@ -11,6 +11,8 @@
 /* define if you have Motif */
 #undef HAVE_MOTIF
 
+#undef HAVE_LIBXAW
+
 #undef TK
 
 @BOTTOM@
@@ -33,7 +35,7 @@
 #ifndef TK
 # if defined HAVE_MOTIF
 #  define TK TK_MOTIF
-# elif defined HAVE_XAW
+# elif defined HAVE_LIBXAW
 #  define TK TK_XAW
 # else
 #  define TK TK_NONE
