@@ -1,6 +1,6 @@
 /*
   xrus - keyboard switcher/indicator
-  Copyright (c) 1996 Alexander V. Lukyanov
+  Copyright (c) 1995-1998 Alexander V. Lukyanov
   This is free software with no warranty.
   See COPYING for details.
 */
@@ -50,6 +50,10 @@
 #define SA_RESTART 0
 #endif
 
+#ifndef R_OK
+# define R_OK 4
+#endif
+
 #include "xalloca.h"
 
 #include "keycomb.h"
@@ -58,6 +62,8 @@
 #include "menu.h"
 #include "kbdstate.h"
 
+#include "confpaths.h"
+
 const char  AppClass[]="Xrus";
 
 extern const char *DefaultResources;
@@ -79,12 +85,14 @@ XrusRec  AppData;
 char  DefaultSwitchKeys[]="Shift_L+Shift_R";
 char  DefaultToLatKeys[]="";
 char  DefaultToRusKeys[]="";
+char  DefaultSwitchForOneKeys[]="";
 char  DefaultLocker[]="exec xlock -remote >/dev/null 2>&1";
 
 char  LockerData[256];
 char  SwitchKeysData[256];
 char  ToLatKeysData[256];
 char  ToRusKeysData[256];
+char  SwitchForOneKeysData[256];
 
 XtResource        resources[]=
 {
@@ -115,6 +123,9 @@ XtResource        resources[]=
    {  "toLatKeys","ToLatKeys",XtRString,  sizeof(String),
          XtOffsetOf(XrusRec,toLatKeys),
          XtRString,  DefaultToLatKeys           },
+   {  "switchForOneKeys","SwitchForOneKeys",XtRString,sizeof(String),
+         XtOffsetOf(XrusRec,switchForOneKeys),
+         XtRString,  DefaultSwitchForOneKeys    },
    {  "xmodmap",  "Xmodmap",  XtRString,  sizeof(String),
          XtOffsetOf(XrusRec,xmodmap),
          XtRString,  ""                         },
@@ -148,6 +159,15 @@ XtResource        resources[]=
    {  "noFork",   "NoFork",   XtRBoolean, sizeof(Boolean),
          XtOffsetOf(XrusRec,noFork),
          XtRImmediate,  (XtPointer)False        },
+   {  "wmIcon",   "WMIcon",   XtRBoolean, sizeof(Boolean),
+            XtOffsetOf(XrusRec,wm_icon),
+            XtRImmediate,  (XtPointer)False     },
+   {  "wMaker",   "WMaker",   XtRBoolean, sizeof(Boolean),
+            XtOffsetOf(XrusRec,wmaker_icon),
+            XtRImmediate,  (XtPointer)False     },
+   {  "xmodmapProgram", "XmodmapProgram", XtRString, sizeof(String),
+            XtOffsetOf(XrusRec,xmodmap_program),
+            XtRString,  BINDIR "/xrusmodmap"     },
 };
 
 XrmOptionDescRec  options[]=
@@ -168,6 +188,8 @@ XrmOptionDescRec  options[]=
    {  "+nofork",    "*noFork",     XrmoptionNoArg,   "false"},
    {  "-fork",      "*noFork",     XrmoptionNoArg,   "false"},
    {  "+fork",      "*noFork",     XrmoptionNoArg,   "true" },
+   {  "-wmicon",    "*wmIcon",     XrmoptionNoArg,   "true"},
+   {  "-wmaker",    "*wMaker",     XrmoptionNoArg,   "true"},
 };
 
 Display        *disp;
@@ -181,7 +203,7 @@ int   count;
 
 char  *program;
 
-struct KeyCombination   SwitchKeys,ToLatKeys,ToRusKeys;
+struct KeyCombination   SwitchKeys,ToLatKeys,ToRusKeys,SwitchForOneKeys;
 
 #define ShouldUseLat()        IsModifierPressed(ControlMapIndex)
 
@@ -191,7 +213,8 @@ enum
    RUS=1,
    LAT=0,
    TEMP_LAT=2,
-   CAPSLOCK_ON=4
+   CAPSLOCK_ON=4,
+   FOR_ONE=8,
 };
 int   Mode=LAT;
 int   NewMode;
@@ -207,8 +230,6 @@ int   format_ret;
 unsigned long nitems_ret,bytes_after_ret;
 unsigned char *prop;
 
-Atom  wm_delete_window;
-
 int   xrus_check=1;
 
 void  AddChildren(Window);
@@ -218,6 +239,8 @@ int   (*old_error_handler)(Display *d,XErrorEvent *ev);
 
 pid_t LockerRunning=0;
 
+time_t alarm_expected;
+
 int   MappingNotifyIgnoreCount=0;
 XtIntervalId   MappingNotifyTimeout=-1;
 XtIntervalId   raise_tmout=-1;
@@ -235,11 +258,39 @@ static void  ShowSwitchButton(void);
 /*________________________________________________________________________
 */
 
+
+/* SwappableKey, modified by Tobias Ernst:
+
+   Now returns 0 if this key is not to be swapped, non-zero otherwise. The
+   non-zero number indicates how many keysyms are grouped into one block. These
+   are two in a standard English-Russian setup, but could be four e.g. in a
+   German-Russian setup (keysyms 0 and 1 for Latin letters plus keysyms 2 and 3
+   for "AltGr" key combinations in a group, and keysyms 4 and 5 for cyrillic
+   letters and 6 and 7 with NoSymbol in the other group.) */
+
 static int SwappableKey(KeySym *k)
 {
-   return(k[0]>0 && k[1]>0 && k[2]>0 && k[3]>0
-       && k[0]<MaxLetterKeySym && k[1]<MaxLetterKeySym
-       && k[2]<MaxLetterKeySym && k[3]<MaxLetterKeySym);
+    int i, j = 0;
+
+    for (i = 2; i <= (keysyms_per_keycode / 2); i += 2)
+    {
+        if (k[i] > 0 && k[i]<MaxLetterKeySym)
+        {
+            j = i;
+        }
+    }
+
+    for (i = 0; i < 2*j; i++)
+    {
+        if (!(k[i] >= 0 && k[i] < MaxLetterKeySym))
+        {
+            /* nonprintable character found other than NoSymbol found,
+               and we don't want to swap this */
+            return 0;
+        }
+    }
+
+    return j;
 }
 
 static int CapsLockableKey(KeySym *k)
@@ -262,7 +313,7 @@ void  SwitchKeyboard(int to)
 {
    KeySym   tmp;
    KeySym   *keymap;
-   int      i;
+   int      i, j, groupwidth;
    XKeyboardControl  kbd_control;
    int   from_mode,to_mode,from_caps,to_caps;
    int   changed=0;
@@ -271,28 +322,34 @@ void  SwitchKeyboard(int to)
       return;
 
    keymap=GetKeymap();
+   if(keymap==0)
+      return;
 
    if(keysyms_per_keycode<4)   /* should not happen */
       return;
 
-   from_mode=((Mode&TEMP_LAT)?LAT:Mode&MODE);
-   to_mode=((to&TEMP_LAT)?LAT:to&MODE);
+   from_mode=((Mode&TEMP_LAT)?LAT:(Mode&FOR_ONE)?((Mode&MODE)^MODE):Mode&MODE);
+   to_mode=  ((to  &TEMP_LAT)?LAT:(to  &FOR_ONE)?((to  &MODE)^MODE):to  &MODE);
+
    from_caps=((Mode&TEMP_LAT)?0:Mode&CAPSLOCK_ON);
    to_caps=((to&TEMP_LAT)?0:to&CAPSLOCK_ON);
 
    for(i=(max_keycode-min_keycode)*keysyms_per_keycode;
        i>=0; i-=keysyms_per_keycode)
    {
-      if(from_mode!=to_mode && SwappableKey(keymap+i))
+      groupwidth = SwappableKey(keymap + i);
+
+      if(from_mode!=to_mode && groupwidth!=0)
       {
-         tmp=keymap[i];
-         keymap[i]=keymap[i+2];
-         keymap[i+2]=tmp;
-         tmp=keymap[i+1];
-         keymap[i+1]=keymap[i+3];
-         keymap[i+3]=tmp;
+         for (j=0; j < groupwidth; j++)
+         {
+             tmp=keymap[i+j];
+             keymap[i+j]=keymap[i+j+groupwidth];
+             keymap[i+j+groupwidth]=tmp;
+         }
          changed=1;
       }
+
       if(AppData.capsLockEmu && from_caps!=to_caps)
       {
          if(CapsLockableKey(keymap+i))
@@ -302,11 +359,11 @@ void  SwitchKeyboard(int to)
             keymap[i+1]=tmp;
             changed=1;
          }
-         if(CapsLockableKey(keymap+i+2))
+         if(groupwidth && CapsLockableKey(keymap+i+groupwidth))
          {
-            tmp=keymap[i+2];
-            keymap[i+2]=keymap[i+3];
-            keymap[i+3]=tmp;
+            tmp=keymap[i+groupwidth];
+            keymap[i+groupwidth]=keymap[i+groupwidth+1];
+            keymap[i+groupwidth+1]=tmp;
             changed=1;
          }
       }
@@ -375,6 +432,15 @@ static void InitCapsLockEmu()
    XFreeModifiermap(mod);
 }
 
+static const char *load_map=0;
+void load_map_delayed(const char *map)
+{
+   if(MappingNotifyIgnoreCount==0)
+      run_xmodmap(map);
+   else
+      load_map=map;
+}
+
 #if TK!=TK_NONE
 void  RaiseButton()
 {
@@ -400,6 +466,7 @@ void  ShowSwitchButton(void)
       to_manage=1;
    else if((Mode&MODE)==LAT && !XtIsManaged(switch_button[0]))
       to_manage=0;
+
    if(to_manage!=-1)
    {
       XtSetArg(args[0],XtNallowShellResize,False);
@@ -419,10 +486,18 @@ void  ShowSwitchButton(void)
 void  UnmapHandler(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
 {
    *cont=True;
-   if(ev->type==UnmapNotify && AppData.alwaysMapped && AppData.icon)
+   if(ev->type==UnmapNotify && AppData.alwaysMapped && AppData.icon
+   && !AppData.wm_icon && !AppData.wmaker_icon)
    {
       XtMapWidget(top_level);
    }
+   else if(ev->type==MapNotify)
+   {
+      if(AppData.wmaker_icon)
+         XWithdrawWindow(disp,XtWindow(top_level),DefaultScreen(disp));
+      else if(AppData.wm_icon)
+         XIconifyWindow(disp,XtWindow(top_level),DefaultScreen(disp));
+   }
 }
 
 void  VisibilityChange(Widget w,XtPointer closure,XEvent *ev,Boolean *cont)
@@ -513,9 +588,19 @@ int   is_wm_window(Window w)
    return is;
 }
 
-void  AddWindow(Window w)
+void  AddWindow(Window w,int tmout);
+
+void  AddWindowOnTimeout(XtPointer closure,XtIntervalId *id)
+{
+   (void)id;
+   DelWindow((Window)closure);
+   AddWindow((Window)closure,False);
+}
+
+void  AddWindow(Window w,int tmout)
 {
    XWindowAttributes wa;
+   long mask;
 
    if(AddToWindowCache(w)) /* if the window was not in cache */
    {
@@ -530,8 +615,11 @@ void  AddWindow(Window w)
          DelWindow(w);
          return;
       }
+
+      mask=wa.your_event_mask;
+
       /* we want to be notified of subwindows creation/destruction etc */
-      wa.your_event_mask|=SubstructureNotifyMask;
+      mask|=SubstructureNotifyMask;
 
       /* we request events only if
          1. the window already requested this kind of events
@@ -539,24 +627,33 @@ void  AddWindow(Window w)
          3. this window is wm's one or top level one
             (for the case the application itself doesn't want these events)
       */
-      wa.your_event_mask|=(wa.all_event_masks|wa.do_not_propagate_mask)
-                          &(KeyPressMask|KeyReleaseMask|PointerMotionMask);
+      mask|=(wa.all_event_masks|wa.do_not_propagate_mask)
+            &(KeyPressMask|KeyReleaseMask|PointerMotionMask);
 
       /* we always want both Press and Release events */
-      if(wa.your_event_mask&(KeyPressMask|KeyReleaseMask))
-         wa.your_event_mask|=(KeyPressMask|KeyReleaseMask);
+      if(mask&(KeyPressMask|KeyReleaseMask))
+         mask|=(KeyPressMask|KeyReleaseMask);
 
       /* third condition */
-      if((wa.your_event_mask&(KeyPressMask|KeyReleaseMask|PointerMotionMask))
-         !=(KeyPressMask|KeyReleaseMask|PointerMotionMask)
+      if((mask&(KeyPressMask|KeyReleaseMask))
+             !=(KeyPressMask|KeyReleaseMask)
       && (w==wa.root || window_is_top_level(w) || is_wm_window(w)))
-         wa.your_event_mask|=(KeyPressMask|KeyReleaseMask|PointerMotionMask);
+         mask|=(KeyPressMask|KeyReleaseMask);
       else
          DelWindow(w);  /* we'll check input mask later */
 
-      XSelectInput(disp,w,wa.your_event_mask);
+      if(w==wa.root)
+	 mask|=PointerMotionMask;   /* always want it for root window */
+
+      XSelectInput(disp,w,mask);
 
       AddChildren(w);
+
+      if(tmout)
+      {
+         XtAppAddTimeOut(app_context,AppData.recheckTime,AddWindowOnTimeout,
+                         (XtPointer)w);
+      }
    }
 }
 
@@ -583,48 +680,54 @@ void  AddChildren(Window w)
    if(XQueryTree(disp,w,&root,&parent,&children,&children_num))
    {
       for(i=0; i<children_num; i++)
-         AddWindow(children[i]);
+         AddWindow(children[i],False);
       XFree(children);
    }
 }
 
-void  AddWindowOnTimeout(XtPointer closure,XtIntervalId *id)
-{
-   (void)id;
-   DelWindow((Window)closure);
-   AddWindow((Window)closure);
-}
-
 void  SwitchKeysFire(void)
 {
    NewMode^=MODE;
+   NewMode&=~FOR_ONE;
 }
 void  ToLatKeysFire(void)
 {
    NewMode&=~MODE;
+   NewMode&=~FOR_ONE;
 }
 void  ToRusKeysFire(void)
 {
    NewMode|=MODE;
+   NewMode&=~FOR_ONE;
+}
+void  SwitchForOneKeysFire(void)
+{
+   NewMode^=FOR_ONE;
 }
 
 void  CheckKeymap()
 {
-   static   first_time=1;
+   static int first_time=1;
 
    int      ok_rus=0;
    int      ok_lat=0;
-   int      i;
+   int      i,groupwidth;
    KeySym   *keymap=GetKeymap();
 
+   if(keymap==0)
+   {
+      Mode=LAT;
+      exit(1);
+   }
+
    if(keysyms_per_keycode>=4)
    {
       for(i=(max_keycode-min_keycode)*keysyms_per_keycode;
           i>=0; i-=keysyms_per_keycode)
       {
-         if(keymap[i] && keymap[i+1] && keymap[i+2] && keymap[i+3])
+         if((groupwidth=SwappableKey(keymap+i)) > 0)
          {
-            ok_lat+=(NationalKeySym(keymap[i+2]) && NationalKeySym(keymap[i+3]));
+            ok_lat+=(NationalKeySym(keymap[i+groupwidth]) && NationalKeySym(keymap[i+groupwidth+1]));
             ok_rus+=(NationalKeySym(keymap[i+0]) && NationalKeySym(keymap[i+1]));
          }
       }
@@ -635,22 +738,22 @@ void  CheckKeymap()
       fprintf(stderr,"%s: key mapping is not in proper coding, exiting.\n",program);
       if(first_time)
       {
-         fprintf(stderr,"   You might probably want to load a localized modmap before xrus starting.\n"
+         fprintf(stderr,"   You might probably want to load a localized modmap before starting xrus.\n"
                         "   Alternatively, you can specify modmap file name on the command line,\n"
-                        "   like this:   ./xrus jcuken-koi8-xrus.xmm\n");
+                        "   like this:   ./xrus jcuken-koi8.xmm\n");
       }
       else
       {
          fprintf(stderr,"   This means your modmap has changed to single-mode one.\n"
                         "   If you were doing xmodmap <file> at the time, check your mapping in\n"
-                        "   the file. For normal xrus work it is required modmap to have both\n"
+                        "   the file. For normal xrus work it is required that modmap has both\n"
                         "   national and latin letters\n");
       }
       Mode=LAT;
       exit(1);
    }
 
-   NewMode=Mode=(Mode&~MODE)|(ok_rus>ok_lat?RUS:LAT);
+   NewMode=Mode=(Mode&~(MODE))|(ok_rus>ok_lat?RUS:LAT);
    FixNewMode();
    SwitchKeyboard(NewMode);
 
@@ -671,8 +774,10 @@ void  MainLoop(void)
 {
    XEvent   ev;
    KeySym   ks;
-   int      i;
    int      was_caps_locked;
+   Bool     fired;
+   int      len;
+   char     string[64];
 
    for(;;)
    {
@@ -691,8 +796,11 @@ void  MainLoop(void)
       switch(ev.type)
       {
       case(MapNotify):
-         AddWindow(ev.xcreatewindow.window);
-         XtAppAddTimeOut(app_context,AppData.recheckTime,AddWindowOnTimeout,(XtPointer)ev.xmap.window);
+         AddWindow(ev.xmap.window,True);
+         break;
+      case(ReparentNotify):
+         /* the structure could change under us - re-add the window */
+         AddWindow(ev.xreparent.window,True);
          break;
       case(DestroyNotify):
          DelWindow(ev.xdestroywindow.window);
@@ -701,6 +809,7 @@ void  MainLoop(void)
          SetAlarm();
 
          ks=XLookupKeysym(&ev.xkey,0);
+         len=XLookupString(&ev.xkey,string,sizeof(string),0,0);
 
          NewMode=Mode;
 
@@ -710,6 +819,10 @@ void  MainLoop(void)
          KeyCombinationProcessKeyPress(&SwitchKeys,ks);
          KeyCombinationProcessKeyPress(&ToRusKeys,ks);
          KeyCombinationProcessKeyPress(&ToLatKeys,ks);
+         KeyCombinationProcessKeyPress(&SwitchForOneKeys,ks);
+
+         if(len>0)
+            NewMode&=~FOR_ONE;
 
          FixNewMode();
          SwitchKeyboard(NewMode);
@@ -722,9 +835,11 @@ void  MainLoop(void)
 
          NewMode=Mode;
 
-         i=KeyCombinationProcessKeyRelease(&SwitchKeys,ks,1);
-         i|=KeyCombinationProcessKeyRelease(&ToLatKeys,ks,!i);
-         KeyCombinationProcessKeyRelease(&ToRusKeys,ks,!i);
+         fired=False;
+         fired|=KeyCombinationProcessKeyRelease(&SwitchKeys,ks,!fired);
+         fired|=KeyCombinationProcessKeyRelease(&ToLatKeys,ks,!fired);
+         fired|=KeyCombinationProcessKeyRelease(&ToRusKeys,ks,!fired);
+         fired|=KeyCombinationProcessKeyRelease(&SwitchForOneKeys,ks,!fired);
 
          FixNewMode();
          SwitchKeyboard(NewMode);
@@ -747,7 +862,14 @@ void  MainLoop(void)
                                        MappingNotifyTimeoutHandler,NULL);
             }
             else
+            {
                MappingNotifyIgnoreCount--;
+               if(MappingNotifyIgnoreCount==0 && load_map!=0)
+               {
+                  run_xmodmap(load_map);
+                  load_map=0;
+               }
+            }
          }
          break;
       }
@@ -806,6 +928,7 @@ void  LockScreen()
 void  run_xmodmap(const char *file)
 {
    char *alt_file;
+   char *xmodmap=AppData.xmodmap_program;
 
    XSync(disp,False);
 
@@ -845,8 +968,8 @@ try_again:
       break;
    case(0):
       /* child */
-      execlp("xmodmap","xmodmap",file,NULL);
-      fprintf(stderr,"%s: execlp(xmodmap) failed - %s\n",program,strerror(errno));
+      execlp(xmodmap,xmodmap,file,NULL);
+      fprintf(stderr,"%s: execlp(%s) failed - %s\n",program,xmodmap,strerror(errno));
       fflush(stderr);
       _exit(1);
    default:
@@ -857,17 +980,34 @@ try_again:
 
 void  AlarmHandler(int sig)
 {
-   (void)sig;
+   time_t now=time(0);
+   if(now < alarm_expected)
+   {
+      fprintf(stderr,"%s: unexpected early alarm signal\n",program);
+      alarm(alarm_expected-now);
+      return;
+   }
    if(AppData.autolock)
+   {
+#ifdef DEBUG
+      printf("calling LockScreen() at %ld\n",now);
+#endif
       LockScreen();
+   }
+   (void)sig;
 }
 
 void  SetAlarm()
 {
+   alarm(0);
    if(AppData.autolock && AppData.timeout>0 && !LockerRunning)
+   {
       alarm(AppData.timeout*60);
-   else
-      alarm(0);
+      alarm_expected=time(0)+AppData.timeout*60-1;
+#ifdef DEBUG
+      printf("alarm expected at %ld\n",alarm_expected);
+#endif
+   }
 }
 
 int   X_error_handler(Display *d,XErrorEvent *ev)
@@ -912,10 +1052,14 @@ int   main(int argc,char **argv)
    char  err[1024];
    int   scr;
 
-#if 0
-   printf("Xrus | Version " VERSION " | Copyright (c) 1995-96 Alexander V. Lukyanov\n");
-   fflush(stdout);
-#endif
+   Window w;
+   Atom  wm_delete_window;
+   Atom  wm_client_leader;
+
+   int   saved_argc=argc;
+   char  **saved_argv=alloca((argc+1)*sizeof(char*));
+   memcpy(saved_argv,argv,argc*sizeof(char*));
+   saved_argv[argc]=0;
 
    program=argv[0];
 
@@ -948,6 +1092,8 @@ int   main(int argc,char **argv)
                 " -led led_no            use the led to indicate alternative mode\n"
                 " +led                   don't use led for indication\n"
                 " -nofork                don't fork to background\n"
+                " -wmaker                display in Window Maker appicon\n"
+                " -wmicon                display in window manager icon\n"
                 "\n",program);
          exit(0);
       }
@@ -987,7 +1133,7 @@ int   main(int argc,char **argv)
    /* add all windows located on the display to the pool
       and check for other xrus presence */
    for(scr=0; scr<ScreenCount(disp); scr++)
-      AddWindow(RootWindowOfScreen(ScreenOfDisplay(disp,scr)));
+      AddWindow(RootWindowOfScreen(ScreenOfDisplay(disp,scr)),True);
    xrus_check=0;
 
    StartArgs();
@@ -1000,12 +1146,13 @@ int   main(int argc,char **argv)
    AddArg(XtNinput,False);
    AddArg(XtNallowShellResize,True);
    AddArg(XtNmappedWhenManaged,False);
-   AddArg(XtNargc,argc);
-   AddArg(XtNargv,argv);
+   AddArg(XtNargc,saved_argc);
+   AddArg(XtNargv,saved_argv);
    top_level=XtAppCreateShell("xrus",AppClass,applicationShellWidgetClass,disp,args,count);
 
    XtGetApplicationResources(top_level,&AppData,resources,XtNumber(resources),NULL,0);
 
+   /* replace pointer to constant string with pointer to our buffers */
    strcpy(LockerData,AppData.locker?AppData.locker:DefaultLocker);
    AppData.locker=LockerData;
    strcpy(SwitchKeysData,AppData.switchKeys?AppData.switchKeys:DefaultSwitchKeys);
@@ -1014,6 +1161,8 @@ int   main(int argc,char **argv)
    AppData.toLatKeys=ToLatKeysData;
    strcpy(ToRusKeysData,AppData.toRusKeys?AppData.toRusKeys:DefaultToRusKeys);
    AppData.toRusKeys=ToRusKeysData;
+   strcpy(SwitchForOneKeysData,AppData.switchForOneKeys?AppData.switchForOneKeys:DefaultSwitchForOneKeys);
+   AppData.switchForOneKeys=SwitchForOneKeysData;
 
    if(argc>1)
    {
@@ -1039,9 +1188,12 @@ int   main(int argc,char **argv)
    ToRusKeys.Fire=ToRusKeysFire;
    ParseKeyCombination(AppData.toRusKeys,&ToRusKeys,err);
    PrintParseErrors("toRusKeys",err);
+   SwitchForOneKeys.Fire=SwitchForOneKeysFire;
+   ParseKeyCombination(AppData.switchForOneKeys,&SwitchForOneKeys,err);
+   PrintParseErrors("switchForOneKeys",err);
 
 #if !defined(DEBUG) && !defined(DONT_FORK)
-   if(!AppData.noFork)
+   if(!AppData.noFork && !AppData.wmaker_icon)
    {
       switch(fork())
       {
@@ -1101,6 +1253,7 @@ int   main(int argc,char **argv)
    ol_decor_del[0]=XInternAtom(disp,"_OL_DECOR_HEADER",False);
    ol_decor_del[1]=XInternAtom(disp,"_OL_DECOR_RESIZE",False);
 
+   wm_client_leader=XInternAtom(disp,"WM_CLIENT_LEADER",False);
    wm_delete_window=XInternAtom(disp,"WM_DELETE_WINDOW",False);
 
    XtRealizeWidget(top_level);
@@ -1108,6 +1261,10 @@ int   main(int argc,char **argv)
    XChangeProperty(disp,XtWindow(top_level),ol_decor_del_atom,XA_ATOM,32,
          PropModeReplace,(void*)ol_decor_del,2);
 
+   w=XtWindow(top_level);
+   XChangeProperty(disp,XtWindow(top_level),wm_client_leader,XA_WINDOW,32,
+         PropModeReplace,(void*)&w,1);
+
 #if TK!=TK_MOTIF && TK!=TK_NONE
 {
    /* we have to emulate motif hints as we don't have Motif */
@@ -1125,9 +1282,6 @@ int   main(int argc,char **argv)
    InitCapsLockEmu();
 
 #if TK!=TK_NONE
-   if(AppData.icon)
-      XtMapWidget(top_level);
-
    if(AppData.adjustModeButtons)
    {
       Dimension w0,h0,w1,h1;
@@ -1154,6 +1308,25 @@ int   main(int argc,char **argv)
       XtSetValues(switch_button[1],args,count);
       XtSetValues(form_w,args,count);
    }
+
+   if(AppData.wm_icon || AppData.wmaker_icon)
+   {
+      int state=IconicState;
+      if(AppData.wmaker_icon)
+         state=WithdrawnState;
+      XtUnmapWidget(form_w);
+      XReparentWindow(disp,XtWindow(form_w),RootWindowOfScreen(ScreenOfDisplay(disp,scr)),0,0);
+      XtVaSetValues(top_level,
+         XtNiconWindow,XtWindow(form_w),
+         XtNinitialState,state,
+         NULL);
+      XtMapWidget(top_level);
+   }
+   else
+   {
+      if(AppData.icon)
+         XtMapWidget(top_level);
+   }
 #endif /* TK!=TK_NONE */
 
    SetAlarm();
