@@ -19,6 +19,7 @@ typedef struct
    String   switchKeys;
    String   toRusKeys;
    String   toLatKeys;
+   String   switchForOneKeys;
    String   alternateMappings;
    String   xmodmap;
    int      led;
@@ -31,17 +32,18 @@ typedef struct
    int      capsLockLed;
    int      recheckTime;
    Boolean  noFork;
+   Boolean  wm_icon;
+   Boolean  wmaker_icon;
+   String   xmodmap_program;
 } XrusRec;
 
 extern   XrusRec  AppData;
 
-extern   char  LockerData[256];
-extern   char  SwitchKeysData[256];
-
 extern   char  DefaultLocker[];
 extern   char  DefaultSwitchKeys[];
 extern   char  DefaultToLatKeys[];
 extern   char  DefaultToRusKeys[];
+extern   char  DefaultSwitchForOneKeys[];
 
 extern   Display        *disp;
 extern   XtAppContext   app_context;
@@ -57,10 +59,14 @@ extern   Widget   switch_button[2];
 void  SetAlarm();
 void  LockScreen();
 void  run_xmodmap(const char *file);
+void  load_map_delayed(const char *map);
+
+void  SwitchKeyboard(int);
+void  FixNewMode();
 
 void  cleanup();  /* should be called before DestroyApplicationContext */
 
-extern struct KeyCombination   SwitchKeys,ToLatKeys,ToRusKeys;
+extern struct KeyCombination   SwitchKeys,ToLatKeys,ToRusKeys,SwitchForOneKeys;
 
 extern   int   Mode,NewMode;
 
